from typing import Callable, Optional, TypeVar, Iterable, Generator
import os
from os import DirEntry
from getpass import getuser as __username
from collections import defaultdict as __defaultdict
import json as __json
from . import constants


T = TypeVar("T")

FilePredicate = Callable[[DirEntry], bool]

Path = str | DirEntry


def __generator(__iter: Iterable[T]) -> Generator[T, None, None]:
    for i in __iter:
        yield i


def __reversed_generator(__iter: Iterable[T]) -> Generator[T, None, None]:
    for i in reversed(__iter):
        yield i


def __filehandler(
    __where: Optional[FilePredicate], __key: Optional[Callable[[DirEntry], T]]
):

    if __where is None:
        if __key is None:

            def file_handler(__file: DirEntry):
                return __file

        else:

            def file_handler(__file: DirEntry):
                return (__file, __key(__file))

    else:
        if __key is None:

            def file_handler(__file: DirEntry):
                return __file if __where(__file) else None

        else:

            def file_handler(__file: DirEntry):
                return (__file, __key(__file)) if __where(__file) else None

    return file_handler


def __recursive_scanner(__abspath: bool, __filehandler: Callable[[DirEntry], T]):
    for file in os.scandir(__abspath):
        if file.is_file():
            value = __filehandler(file)
            if value is not None:
                yield value
        elif file.is_dir():
            for result in __recursive_scanner(file.path, __filehandler):
                yield result


def __scandir(
    __path: str,
    __is_abspath: bool,
    __where: Optional[FilePredicate],
    __key: Optional[Callable[[DirEntry], T]],
    __reverse: bool,
    __recursive: bool,
):
    path = __abspath(__path, __is_abspath)
    if __recursive:
        result = __generator(__recursive_scanner(path, __filehandler(__where, __key)))
    else:
        result = (file for file in os.scandir(path))
        if __where is not None:
            result = (file for file in result if __where(file))
        if __key is not None:
            result = list(result)
            result.sort(key=__key)
            result = __generator(result)

    if __reverse:
        result = __reversed_generator(result)

    return result


def __countdir(
    __path: Path,
    __is_abspath: bool,
    __where: Optional[FilePredicate],
    __recursive: bool,
):
    n = 0
    for _ in __scandir(__path, __is_abspath, __where, None, False, __recursive):
        n += 1
    return n


def __remove_path(__path: Path):
    path = __path.path if isinstance(__path, DirEntry) else __path
    if __exists(path, True):
        os.remove(path)


def __remove(__path: str, __is_abspath: bool, __where: Optional[FilePredicate]):
    for file in __scandir(__path, __is_abspath, __where, None, False):
        __remove_path(file)


def __system(__command: str):
    return os.system(command=__command)


def __makedirs(__name: str, __is_abspath: bool, __exist_ok: bool):
    path = __abspath(__name, __is_abspath)
    if not __dir_exists(path, True):
        os.makedirs(path, exist_ok=__exist_ok)


def __mkdir(__name: str, __is_abspath: bool, __exist_ok: bool):
    return __makedirs(__name, __is_abspath, __exist_ok)


def __mkfile(__name: str, __is_abspath: bool, __exist_ok: bool):
    path = __abspath(__name, __is_abspath)
    if not __file_exists(path, True):
        if __exist_ok:
            pdir = __parentdir(path, True)
            if not __dir_exists(pdir, True):
                __makedirs(pdir, True, __exist_ok)
            __writes(path, "", True, constants.ERRORS, constants.ENCODING, True)


def __filetree_paths(__ft: str, __parentdir: Optional[str], __indent: int) -> list[str]:
    paths, dirlog = [], __defaultdict(list[str])
    for line in (line for line in __ft.splitlines(keepends=False) if line):
        offset, path = len(line) - len(line.lstrip()), line.strip()
        dirlog[offset].append(path)
        while offset > 0:
            if dirlog[offset - __indent]:
                path = f"{dirlog[offset-__indent][-1]}/{path}"
            offset -= __indent
        if path:
            paths.append(path)
    if __parentdir is not None and __parentdir:
        for (i, path) in (
            (i, p) for (i, p) in enumerate(paths) if len(p) - len(p.lstrip()) == 0
        ):
            paths[i] = __join(__parentdir, path)
    return paths


def __mkfiletree(
    __ft: str, __overwrite_if_exists: bool, __parentdir: Optional[str], __indent: int
):
    ft_paths = __filetree_paths(__ft, __parentdir, __indent)
    dirpaths = set(
        path
        for path in ft_paths
        if (
            next(
                (
                    ft_path
                    for ft_path in ft_paths
                    if (ft_path.startswith(f"{path}/") and ft_path != path)
                ),
                None,
            )
            is not None
        )
        and not "." in __basename(path)
    )

    for dirpath in dirpaths:
        __mkdir(dirpath, False, True)

    for path in (path for path in ft_paths if not path in dirpaths):
        if __file_exists(path, False):
            if __overwrite_if_exists:
                __writes(path, "", False, constants.ERRORS, constants.ENCODING, True)
        else:
            __mkfile(path, False, True)

    for path in ft_paths:
        if not __exists(path, False):
            print(f"error: path does not exist:{path}")


def __ext(__path: Path):
    path = __path.name if isinstance(__path, DirEntry) else __path
    return path.split(".")[-1] if "." in path else path


def __join(__a: str, *paths):
    return os.path.join(__a, *paths)


def __basename(__name: str):
    return __name.split("/")[-1] if "/" in __name else __name


def __parentdir(__name: str, __is_abspath: bool):
    result = __abspath(__name, __is_abspath)
    return "/".join(result.split("/")[:-1]) if "/" in result else result


def __isabs(__path: str):
    return os.path.isabs(__path)


def __isfile(__path: str):
    return os.path.isfile(__path)


def __isdir(__path: str):
    return os.path.isdir(__path)


def __islink(__path: str):
    return os.path.islink(__path)


def __safeabs(__name: str, __is_abspath: bool):
    return __name if __is_abspath and __isabs(__name) else __abspath(__name, False)


def __safepath(__name: str, __is_abspath: bool, __exist_ok: bool):
    path = __abspath(__name, __is_abspath)
    dirname = __parentdir(path, True)
    if not __dir_exists(dirname, True):
        os.makedirs(dirname, __exist_ok)
    return path


def __exists(__name: str, __is_abspath: bool):
    return os.path.exists(__safeabs(__name, __is_abspath))


def __file_exists(__name: str, __is_abspath: bool):
    path = __safeabs(__name, __is_abspath)
    return __exists(path, True) and __isfile(path)


def __dir_exists(__name: str, __is_abspath: bool):
    path = __safeabs(__name, __is_abspath)
    return __exists(path, True) and __isdir(path)


def __abspath(__path: str, __is_abspath: bool):
    return __path if __is_abspath and __isabs(__path) else os.path.abspath(__path)


def __ctime(__path: str):
    return os.path.getctime(__path)


def __atime(__path: str):
    return os.path.getatime(__path)


def __mtime(__path: str):
    return os.path.getmtime(__path)


def __size(__path: str):
    return os.path.getsize(__path)


def __writes(
    __name: str,
    __content: str,
    __is_abspath: bool,
    __errors: str,
    __encoding: str,
    __exist_ok: bool,
):
    path = __safepath(__name, __is_abspath, __exist_ok)
    with open(path, "w+", errors=__errors, encoding=__encoding) as file:
        file.write(__content)
        file.close()


def __writeb(
    __name: str,
    __content: bytes,
    __is_abspath: bool,
    __errors: str,
    __encoding: str,
    __exist_ok: bool,
):
    path = __safepath(__name, __is_abspath, __exist_ok)
    with open(path, "w+", errors=__errors, encoding=__encoding) as file:
        file.write(__content)
        file.close()


def __writejson(
    __name: str,
    __content,
    __indent: int,
    __is_abspath: bool,
    __exist_ok: bool,
):
    path = __safepath(__name, __is_abspath, __exist_ok)
    s = __json.dumps(__content, indent=__indent)
    return __writes(path, s, True, constants.ERRORS, constants.ENCODING, __exist_ok)


def __write(
    __name: str,
    __content,
    __indent: int,
    __is_abspath: bool,
    __errors: str,
    __encoding: str,
    __exist_ok: bool,
):
    path = __safepath(__name, __is_abspath, __exist_ok)
    return (
        __writes(path, __content, True, __errors, __encoding, __exist_ok)
        if isinstance(__content, str)
        else __writejson(path, __content, __indent, True, __exist_ok)
    )


def __read_or_none(func):
    def inner(*a, **b):
        try:
            result = func(*a, **b)
        except Exception:
            result = None
        return result

    return inner


@__read_or_none
def __reads(__name: str, __is_abspath: bool):
    path = __abspath(__name, __is_abspath)
    result = None
    with open(path, mode="r") as file:
        result = str(file.read())
    return result


@__read_or_none
def __readb(__name: str, __is_abspath: bool):
    path = __abspath(__name, __is_abspath)
    with open(path, mode="rb") as file:
        result = file.read()
    return result


@__read_or_none
def __readjson(__name: str, __is_abspath: bool):
    return __json.loads(__reads(__name, __is_abspath))


@__read_or_none
def __readlines(
    __name: str,
    __is_abspath: bool,
    __keepends: bool,
    __where: Optional[Callable[[str], bool]],
):
    s = __reads(__name, __is_abspath)
    result = s.splitlines(keepends=__keepends)
    if __where is not None:
        result = [line for line in result if __where(line)]
    return result


def join(__a: str, *paths):
    return __join(__a, *paths)


def isfile(__path: str):
    return __isfile(__path)


def isdir(__path: str):
    return __isdir(__path)


def islink(__path: str):
    return __islink(__path)


def isabs(__path: str):
    return __isabs(__path)


def abspath(__path: str, is_abspath: bool = False):
    return __abspath(__path, is_abspath)


def atime(__path: str):
    """access time; time last accessed"""
    return __atime(__path)


def ctime(__path: str):
    """created time ; time file is created at"""
    return __ctime(__path)


def mtime(__path: str):
    """modified time; time file was last modified"""
    return __mtime(__path)


def size(__path: str):
    """file size"""
    return __size(__path)


def exists(__name: str, is_abspath: bool = False):
    return __exists(__name, is_abspath)


def file_exists(__name: str, is_abspath: bool = False):
    return __file_exists(__name, is_abspath)


def dir_exists(__name: str, is_abspath: bool = False):
    return __dir_exists(__name, is_abspath)


def basename(__name: str):
    return __basename(__name)


def parentdir(__name: str, is_abspath: bool = False):
    return __parentdir(__name, is_abspath)


def writes(
    __name: str,
    content: str,
    is_abspath: bool = False,
    errors: str = constants.ERRORS,
    encoding: str = constants.ENCODING,
    exist_ok: bool = True,
):
    return __writes(__name, content, is_abspath, errors, encoding, exist_ok)


def writeb(
    __name: str,
    content: bytes,
    is_abspath: bool = False,
    errors: str = constants.ERRORS,
    encoding: str = constants.ENCODING,
    exist_ok: bool = True,
):
    return __writeb(__name, content, is_abspath, errors, encoding, exist_ok)


def writejson(
    __name: str,
    content,
    indent: int = 4,
    is_abspath: bool = False,
    exist_ok: bool = True,
):
    return __writejson(__name, content, indent, is_abspath, exist_ok)


def write(
    __name: str,
    content,
    indent: int = 4,
    is_abspath: bool = False,
    errors: str = constants.ERRORS,
    encoding: str = constants.ENCODING,
    exist_ok: bool = True,
):
    return __write(__name, content, indent, is_abspath, errors, encoding, exist_ok)


def reads(__name: str, is_abspath: bool = False):
    return __reads(__name, is_abspath)


def readb(__name: str, is_abspath: bool = False):
    return __readb(__name, is_abspath)


def readjson(__name: str, is_abspath: bool = False):
    return __readjson(__name, is_abspath)


def readlines(
    __name: str,
    is_abspath: bool = False,
    keepends: bool = False,
    where: Optional[Callable[[str], bool]] = None,
):
    return __readlines(__name, is_abspath, keepends, where)


def system(__command: str):
    return __system(__command)


def makedirs(__name: str, is_abspath: bool = False, exist_ok: bool = True):
    return __makedirs(__name, is_abspath, exist_ok)


def mkfile(__name: str, is_abspath: bool = False, exist_ok: bool = True):
    return __mkfile(__name, is_abspath, exist_ok)


def mkdir(__name: str, is_abspath: bool = False, exist_ok: bool = True):
    return __mkdir(__name, is_abspath, exist_ok)


def mkfiletree(
    __ft: str,
    overwrite_if_exists: bool = False,
    parentdir: Optional[str] = None,
    indent: int = 4,
):
    return __mkfiletree(__ft, overwrite_if_exists, parentdir, indent)


def mk_filetree(
    __ft: str,
    overwrite_if_exists: bool = False,
    parentdir: Optional[str] = None,
    indent: int = 4,
):
    """soon to be deprecated"""
    return __mkfiletree(__ft, overwrite_if_exists, parentdir, indent)


def scandir(
    __path: str,
    is_abspath: bool = False,
    where: Optional[FilePredicate] = None,
    key: Optional[Callable[[DirEntry], T]] = None,
    reverse: bool = False,
    recursive: bool = False,
):
    return __scandir(__path, is_abspath, where, key, reverse, recursive)


def countdir(
    __path: Path,
    is_abspath: bool = False,
    where: Optional[FilePredicate] = None,
    recursive: bool = False,
):
    return __countdir(__path, is_abspath, where, recursive)


def remove_path(__path: Path):
    return __remove_path(__path)


def remove(
    __path: str,
    is_abspath: bool = False,
    where: Optional[Callable[[DirEntry], bool]] = None,
):
    return __remove(__path, is_abspath, where)


def username():
    return __username()


def ext(__path: Path):
    return __ext(__path)
